#-------------------------------------------------------------------------------
# Name:        Aggregate Points
# Purpose:     Aggregate points in a large point feature class so that
#              points in a cluster can be displayed using a single symbol
#
# Author:      Barry Kronenfeld
#
# Created:     27/09/2015
# Copyright:   (c) Barry Kronenfeld 2015
# Licence:     GPL 3.0
#-------------------------------------------------------------------------------

import arcpy
import os
import random
#import pip
import numpy
import kdtree
import timeit

##def got_scipy():
##    # returns true if scipy installed
##    pck_list=pip.get_installed_distributions()
##    for pck in pck_list:
##        if 'scipy' in pck.egg_name():
##            return True
##    return False

def report(msg, severity=0):
    print msg
    if severity==0:
        arcpy.AddMessage(msg)
    if severity==1:
        arcpy.AddWarning(msg)
    if severity>1:
        arcpy.AddError(msg)

report("Finished importing necessary libraries.")

def create_out_fc(infcs,outfc):
    # creates output feature class with column for each input
    # returns list of fieldnames
    # or boolean False if unsuccessful
    arcpy.env.overwriteOutput=True
    outtype="gdb"
    if outfc[-3:]=="shp": outtype="shapefile"
    fieldnames=[]
    # obtain feature class names for columns
    for infc in infcs:
        # obtain feature class name
        parsethis=''
        parsethis=infc.split('\\')
        fc=parsethis[-1]
        parsethis=fc.split('.')
        fc=parsethis[0]
        # replace non alpha-numeric characters
        for i in range(len(fc)-1,0,-1):
            if not fc[i].isalnum:
                fc=fc[:i] + '_' + fc[i+1:]
        # trim
        # gdb field names can have max 64 chars
        # shapefile field names can have max 10 characters
        keepto=len(fc)
        if outtype=="shapefile":
            if keepto > 10:
                keepto = 10
        else:
            if keepto > 64:
                keepto=64
        if keepto<len(fc):
            fc=fc[:keepto]
        # check for duplicate fields
        if fc in fieldnames:
            return False
        # append
        fieldnames.append(fc)
    # get workspace & file name
    ws=os.path.dirname(outfc)
    fcbase=os.path.basename(outfc)
    # create output feature class
    outfc2=arcpy.CreateFeatureclass_management(ws,fcbase,"POINT",spatial_reference=infcs[0])
    if str(outfc2) != outfc:
        return False
    # create fields
    existingfields=arcpy.ListFields(outfc)
    existingfieldnames=[]
    for field in existingfields:
        existingfieldnames.append(field.name)
    for i in range(len(fieldnames)):
        fieldname=fieldnames[i]
        if fieldname in existingfieldnames:
            field=existingfields[i]
            if field.type != "Integer" and field.type != "SmallInteger":
                report("Error: Output feature class field {} exists but is not an integer type. Please rename input dataset.".format(fieldname))
        else:
            arcpy.AddField_management(outfc,fieldname,"LONG")
    # report success
    return fieldnames

def getmasterlist(infcs):
    # creates master list of points
    # each point is a list of x,y,clusterid,fcid (0-based integer)
    mlist=[]
    for i in range(len(infcs)):
        fc=infcs[i]
        sc=arcpy.da.SearchCursor(fc,['SHAPE@X','SHAPE@Y'])
        for row in sc:
            mlist.append([row[0],row[1],0,i])
        del row
        del sc
    return mlist

def getrandomid(probabilities):
    # selects a random id
    # given a list of probabilities that each id will be selected
    diceroll = random.uniform(0,1)
    cum_prob=probabilities[0]
    i=0
    while cum_prob < diceroll and i < len(probabilities)-1:
        i+=1
        cum_prob += probabilities[i]
    return i

def getDistancesToClusters(mlist,clist):
    # gets distances from each point in mlist to each point in clist
    r=[[0 for i in range(len(clist))] for i in range(len(mlist))]
    dummy=3
    for i in range(len(mlist)):
        for j in range(len(clist)):
            x0=mlist[i][0]
            y0=mlist[i][1]
            x1=clist[j][0]
            y1=clist[j][1]
            d=(((x0-x1)**2)+((y0-y1)**2))
            r[i][j]=d
    return r

def updateMinD(mlist,clist,minD):
    # updates list of minimum distances by checking distance
    # to last cluster
    # if minD is empty list, will populate it
    d=[0]*len(mlist)
    for i in range(len(mlist)):
        j=len(clist)-1
        x0=mlist[i][0]
        y0=mlist[i][1]
        x1=clist[j][0]
        y1=clist[j][1]
        d[i]=(((x0-x1)**2)+((y0-y1)**2))
    if len(minD)==0:
        minD=d
    else:
        for i in range(len(minD)):
            if d[i]<minD[i]:
                minD[i]=d[i]
    return minD

##def initClusterCenters2(mlist,numclusters):
##    # spreads out clusters by sorting x & y, choosing quantile values
##    pts_per_dim = int((numclusters**0.5)+1)
##    marray=numpy.array(mlist)
##    X=marray[:,0]
##    Y=marray[:,1]
##    cc=numpy.zeros((pts_per_dim**2,2))
##    interval=float(len(mlist))/(pts_per_dim+1)
##    ccnum=0
##    for row in range(pts_per_dim):
##        for col in range(pts_per_dim):
##            xseed=random.random()
##            yseed=random.random()
##            xpos=int(xseed*interval+col*interval)
##            ypos=int(yseed*interval + row*interval)
##            if xpos>=len(mlist):
##                xpos=len(mlist)-1
##            if ypos>=len(mlist):
##                ypos=len(mlist)-1
##            cc[ccnum,0]=X[xpos]
##            cc[ccnum,1]=Y[ypos]
##            ccnum+=1
##    cc=cc[:numclusters,]
##    return cc.tolist()

def initClusterCenters(mlist,numclusters):
    # returns a list of cluster centers
    # derived using k++ algorithm
    selpointid=random.randint(0,numclusters-1)
    selpoint=mlist[selpointid]
    clist=[[selpoint[0],selpoint[1]]]
    minD=[]
    while len(clist) < numclusters:
        # update minimum distances to nearest cluster
        minD=updateMinD(mlist,clist,minD)
        # get probabilities
        sum_minD=sum(minD)
        p=[d/sum_minD for d in minD]
        # select point from master list at random
        selpointid=getrandomid(p)
        selpoint=mlist[selpointid]
        clist.append([selpoint[0],selpoint[1]])
    return clist

def assignPointsToClusters(mlist,clist):
    # returns copy of mlist with cluster assignment
    # as well as average distance to cluster center
    d=getDistancesToClusters(mlist,clist)

##    ctree=kdtree.KDTree.construct_from_data(clist)
    avg_sqD=0
    for i in range(len(mlist)):
##        nn=ctree.query([mlist[i][0],mlist[i][1]])
##        mlist[i][2]=nn[0][0]
##        avg_sqD += nn[0][1]

        dlist=d[i]
        minD=min(dlist)
        index=dlist.index(minD)
        mlist[i][2]=index
        avg_sqD += minD
    avg_sqD=avg_sqD/len(mlist)
    return mlist,avg_sqD**0.5

def recalculateClusterCenters(mlist,clist):
    # returns copy of clist with new coordinates
    # initialize membership counts in each cluster
    mcount=[0 for i in range(len(clist))]
    # reset cluster centers to 0,0
    clist=[[0,0] for i in range(len(clist))]
    # go through members and increment x,y,count
    for i in range(len(mlist)):
        c=mlist[i][2]
        clist[c][0]=clist[c][0]+mlist[i][0]
        clist[c][1]=clist[c][1]+mlist[i][1]
        mcount[c]+=1
    # compute average x,y, checking for zero membership
    for i in range(len(mcount)):
        # check for no members
        if mcount[i]==0:
            # put cluster at location of random point
            ptid=random.randint(0,len(mlist)-1)
            clist[i][0]=mlist[ptid][0]
            clist[i][1]=mlist[ptid][1]
        else: # calculate average
            clist[i][0]=clist[i][0]/mcount[i]
            clist[i][1]=clist[i][1]/mcount[i]
    return clist

def writeClusters(mlist,clist,outfc,fields):
    # writes the results to the output feature class
    # count number of points from each input in each cluster
    mcount=[[0 for i in range(len(fields))] for i in range(len(clist))]
    for i in range(len(mlist)):
        inputnum=mlist[i][3]
        c=mlist[i][2]
        mcount[c][inputnum]+=1
    # append  geometry field
    fields.append('SHAPE@XY')
    # create insert cursor
    ic=arcpy.da.InsertCursor(outfc,fields)
    # go through clusters
    for i in range(len(clist)):
        row=mcount[i]
        row.append((clist[i][0],clist[i][1]))
        ic.insertRow(row)
    # delete cursor
    del ic

def hasduplicates(alist):
    for i in range(len(alist)):
        for j in range(i+1,len(alist)):
            if alist[i]==alist[j]:
                return True
    return False

def aggregatepoints(infcs,outfc,numclusters):
    # set up output
    report("Input:")
    for fc in infcs:
        report(fc)
    report("# clusters: " + str(numclusters))
    report("Creating output feature class...")
    fieldnames=create_out_fc(infcs,outfc)
    if fieldnames==False:
        report("Could not create output feature class. Input feature classes have identical names after parsing. Please make names unique.",2)
    # get coordinates
    report("Obtaining coordinate data...")
    mlist=getmasterlist(infcs)
    # initialize clusters
    report("Initializing clusters...")
    clist=initClusterCenters(mlist,numclusters)
    mlist,avgD=assignPointsToClusters(mlist,clist)
    # iterate
    report("Iterating...")
    itnum=0
    error_threshold=0.01
    retry_attempts=1
    retry_num=0
    improvement=error_threshold+1
    while improvement>error_threshold:
        start=timeit.default_timer()
        # go through cluster recalculation & reassignment
        last_avgD=avgD
        clist=recalculateClusterCenters(mlist,clist)
        mlist,avgD=assignPointsToClusters(mlist,clist)
        improvement=last_avgD-avgD
        # check error
        if improvement<=error_threshold:
            retry_num +=1
            if retry_num <= retry_attempts:
                improvement=error_threshold+1
        itnum+=1
        stop=timeit.default_timer()
        ittime=stop-start
        report("Iteration {} error is {} ({} sec)".format(itnum,avgD,ittime))
    report("Writing results to output feature class...")
    writeClusters(mlist,clist,outfc,fieldnames)
    report("Finished.")

if __name__ == '__main__':
    ArcMap=0
    IDE=1

    run_in = ArcMap

    infcs=''
    outfc=''
    numclusters=0

    if run_in==IDE:
        infc1=r'C:\Box Sync\programming\Twitter\data\Tweets.gdb\Pope'
        infc2=r'C:\Box Sync\programming\Twitter\data\Tweets.gdb\Xi'
        infcs='{};{}'.format(infc1,infc2)
        numclusters=60
        outfc=r'C:\Box Sync\programming\Twitter\data\Tweets.gdb\outtest'
    if run_in==ArcMap:
        infcs=arcpy.GetParameterAsText(0)
        numclusters=arcpy.GetParameter(1)
        outfc=arcpy.GetParameterAsText(2)

    aggregatepoints(infcs.split(';'), outfc,numclusters)
